/*
 * Copyright (c) 2001, 2002 Jason Tishler
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * A copy of the GNU General Public License can be found at
 * http://www.gnu.org/
 *
 * Written by Jason Tishler <jason@tishler.net>
 *
 * $Id: rebase.c,v 1.18 2002/07/03 17:48:08 jt Exp $
 */

#include <stdio.h>
#include <time.h>
#include <stdlib.h>
#include <windows.h>
#include <imagehlp.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include "getopt.h"

void parse_args (int argc, char *argv[]);
unsigned long string_to_ulong (const char *string);
void usage ();
BOOL get_pathname (const char *filename, char *pathname, unsigned int size);
BOOL make_tempname (char *tempname, unsigned int size);
int copy (const char *source, const char *dest);
BOOL is_rebaseable (const char *pathname);
int full_rename (const char *oldpath, const char *newpath);
void version ();

ULONG image_base = 0;
BOOL down_flag = FALSE;
ULONG offset = 0;
int args_index = 0;
int verbose = 0;
char id[] = "$Id: rebase.c,v 1.18 2002/07/03 17:48:08 jt Exp $";

int
main (int argc, char *argv[])
{
  ULONG new_image_base = 0;
  int i = 0;
  char pathname[MAX_PATH + 1];
  char tempname[MAX_PATH + 1];

  parse_args (argc, argv);
  new_image_base = image_base;

  for (i = args_index; i < argc; i++)
    {
      ULONG old_image_size, old_image_base, new_image_size, prev_new_image_base;
      BOOL status, status2;
      int unix_status;

      /* Get the full pathname to prevent ReBaseImage() from using it's stupid
       * search algorithm (e.g, PATH, etc.) */
      get_pathname (argv[i], pathname, sizeof (pathname));

      /* Skip if not rebaseable. */
      if (!is_rebaseable (pathname))
	{
	  if (verbose)
	    fprintf (stderr, "%s: skipped because not rebaseable\n", pathname);
	  continue;
	}

      /* Make copy to guard against ReBaseImage() corruption.*/
      make_tempname (tempname, sizeof (tempname));
      unix_status = copy (pathname, tempname);
      if (unix_status < 0)
	{
	  fprintf (stderr, "cannot copy %s\n", pathname);
	  break;
	}

      /* Calculate next base address, if rebasing down. */
      if (down_flag)
	new_image_base -= offset;

      /* Rebase the image. */
      prev_new_image_base = new_image_base;
      status = ReBaseImage (tempname,		/* CurrentImageName */
			     "",		/* SymbolPath */
			     TRUE,		/* fReBase */
			     FALSE,		/* fRebaseSysfileOk */
			     down_flag,		/* fGoingDown */
			     0,			/* CheckImageSize */
			     &old_image_size,	/* OldImageSize */
			     &old_image_base,	/* OldImageBase */
			     &new_image_size,	/* NewImageSize */
			     &new_image_base,	/* NewImageBase */
			     time (0));		/* TimeStamp */

      /* ReBaseImage seems to never returns false! */
      status2 = GetLastError ();
      if (status2 != 0)
	{
	  fprintf (stderr, "ReBaseImage (%s) failed with last error = %lu\n",
		   pathname, GetLastError ());
	  continue;
	}

      /* Check for ReBaseImage () corruption */
      if (is_rebaseable (tempname))
	{
	  full_rename (tempname, pathname);
	}
      else
	{
	  fprintf (stderr, "%s: skipped to avoid ReBaseImage() corruption\n",
		   pathname);
	  unlink (tempname);
	  continue;
	}

      /* Display rebase results, if verbose. */
      if (verbose)
	{
	  printf ("%s: new base = %lx, new size = %lx\n", pathname,
		  ((down_flag) ? new_image_base : prev_new_image_base),
		  new_image_size + offset);
	}

      /* Calculate next base address, if rebasing up. */
      if (!down_flag)
	new_image_base += offset;
    }

  return 0;
}

void
parse_args (int argc, char *argv[])
{
  const char *anOptions = "b:do:vV";
  int anOption = 0;

  for (anOption; (anOption = getopt (argc, argv, anOptions)) != -1;)
    {
      switch (anOption)
	{
	case 'b':
	  image_base = string_to_ulong (optarg);
	  break;
	case 'd':
	  down_flag = TRUE;
	  break;
	case 'o':
	  offset = string_to_ulong (optarg);
	  break;
	case 'v':
	  verbose = TRUE;
	  break;
	case 'V':
	  version ();
	  exit (1);
	  break;
	default:
	  usage ();
	  exit (1);
	  break;
	}
    }

  if (image_base == 0)
    {
      usage ();
      exit (1);
    }

  args_index = optind;
}

unsigned long
string_to_ulong (const char *string)
{
  unsigned long number = 0;
  number = strtoul (string, 0, 0);
  return number;
}

void
usage ()
{
  fprintf (stderr,
	   "usage: rebase -b BaseAddress [-Vdv] [-o Offset] ImageFileName ...\n");
}

BOOL
get_pathname (const char *filename, char *pathname, unsigned int size)
{
  char *dummy;
  DWORD status = GetFullPathName (filename, size, pathname, &dummy);
  return (status > 0);
}

BOOL
make_tempname (char *tempname, unsigned int size)
{
  const char *dir = ".";
  const char *prefix = ".tmp.";
  char *p = 0;
  BOOL status = FALSE;

  p = tempnam (dir, prefix);
  if (p)
    {
      strncpy (tempname, p, size);
      free (p);
      status = TRUE;
    }
  return status;
}

int
copy (const char *source, const char *dest)
{
  int source_fd = -1;
  int dest_fd = -1;
  int status = 0;
  mode_t mode = 0777;

  source_fd = open (source, O_RDONLY | O_BINARY);
  if (source_fd < 0)
    {
      status = -1;
      goto done;
    }

  dest_fd = open (dest, O_CREAT| O_WRONLY| O_TRUNC | O_BINARY, mode);
  if (dest_fd < 0)
    {
      status = -1;
      goto done;
    }

  for (;;)
    {
      char buffer[1024];
      int n, n2;

      n = read (source_fd, buffer, sizeof (buffer));
      if (n < 0)
	{
	  status = -1;
	  goto done;
	}

      if (n == 0)
	break;

      n2 = write (dest_fd, buffer, n);
      if (n < 0 || n != n2)
	{
	  status = -1;
	  goto done;
	}
    }

done:
  close (source_fd);
  close (dest_fd);
  return status;
}

BOOL
is_rebaseable (const char *pathname)
{
  const int pe_signature_offset = 128;
  const int pe_characteristics_offset = 150;
  BOOL status = FALSE;
  int fd, size;
  long offset;
  DWORD pe_sig;
  WORD pe_char;

  fd = open (pathname, O_RDONLY);
  if (fd == -1)
    goto done;

  offset = lseek (fd, pe_signature_offset, SEEK_CUR);
  if (offset == -1)
    goto done;

  pe_sig = 0;
  size = read (fd, &pe_sig, sizeof (pe_sig));
  if (size != sizeof (pe_sig))
    goto done;

  if (pe_sig != IMAGE_NT_SIGNATURE)
    goto done;

  lseek (fd, 0, SEEK_SET);
  offset = lseek (fd, pe_characteristics_offset, SEEK_CUR);
  if (offset == -1)
    goto done;

  pe_char = 0;
  size = read (fd, &pe_char, sizeof (pe_char));
  if (size != sizeof (pe_char))
    goto done;

  status = ((pe_char & IMAGE_FILE_RELOCS_STRIPPED) == 0) ? TRUE : FALSE;

done:
  close (fd);
  return status;
}

int
full_rename (const char *oldpath, const char *newpath)
{
  int status;

  status = unlink (newpath);
  if (status < 0)
    {
      fprintf (stderr, "unlink (%s) failed with errno = %d\n", newpath, errno);
      return status;
    }

  status = rename (oldpath, newpath);
  if (status < 0)
    {
      fprintf (stderr, "rename (%s, %s) failed with errno = %d\n", oldpath,
	       newpath, errno);
      return status;
    }

  return status;
}

void
version ()
{
  fprintf (stderr, "%s\n", id);
}
